# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from knack.log import get_logger
from knack.util import CLIError

from azure.mgmt.iotcentral.models import (AppSkuInfo,
                                          App)

from ._client_factory import resource_service_factory

logger = get_logger(__name__)


def iotcentral_app_create(
        cmd, client, app_name, resource_group_name, subdomain, sku="ST2",
        location=None, template=None, display_name=None
):
    cli_ctx = cmd.cli_ctx
    location = _ensure_location(cli_ctx, resource_group_name, location)
    display_name = _ensure_display_name(app_name, display_name)
    appSku = AppSkuInfo(name=sku)

    app = App(subdomain=subdomain,
              location=location,
              display_name=display_name,
              sku=appSku,
              template=template)

    createResult = client.apps.create_or_update(
        resource_group_name, app_name, app)
    return createResult


def iotcentral_app_get(client, app_name, resource_group_name=None):
    if resource_group_name is None:
        return _get_iotcentral_app_by_name(client, app_name)
    return client.apps.get(resource_group_name, app_name)


def iotcentral_app_delete(client, app_name, resource_group_name):
    return client.apps.delete(resource_group_name, app_name)


def iotcentral_app_list(client, resource_group_name=None):
    if resource_group_name is None:
        return client.apps.list_by_subscription()
    return client.apps.list_by_resource_group(resource_group_name)


def iotcentral_app_update(client, app_name, parameters, resource_group_name):
    etag = parameters.additional_properties['etag']
    return client.apps.update(resource_group_name, app_name, parameters, {'IF-MATCH': etag})


def _ensure_location(cli_ctx, resource_group_name, location):
    """Check to see if a location was provided. If not,
        fall back to the resource group location.
    :param object cli_ctx: CLI Context
    :param str resource_group_name: Resource group name
    :param str location: Location to create the resource
    """
    if location is None:
        resource_group_client = resource_service_factory(
            cli_ctx).resource_groups
        return resource_group_client.get(resource_group_name).location
    return location


def _ensure_display_name(app_name, display_name):
    if not display_name or display_name.isspace():
        return app_name
    return display_name


def _get_iotcentral_app_by_name(client, app_name):
    """Search the current subscription for an app with the given name.
    :param object client: IoTCentralClient
    :param str app_name: App name to search for
    """
    all_apps = iotcentral_app_list(client)
    if all_apps is None:
        raise CLIError(
            "No IoT Central application found in current subscription.")
    try:
        target_app = next(
            x for x in all_apps if app_name.lower() == x.name.lower())
    except StopIteration:
        raise CLIError(
            "No IoT Central application found with name {} in current subscription.".format(app_name))
    return target_app



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from knack.log import get_logger

logger = get_logger(__name__)


def rest_call(cmd, url, method=None, headers=None, uri_parameters=None,
              body=None, skip_authorization_header=False, resource=None, output_file=None):
    from azure.cli.core.util import send_raw_request
    r = send_raw_request(cmd.cli_ctx, method, url, headers, uri_parameters, body,
                         skip_authorization_header, resource, output_file)
    if not output_file and r.content:
        try:
            return r.json()
        except ValueError:
            logger.warning('Not a json response, outputting to stdout. For binary data '
                           'suggest use "--output-file" to write to a file')
            print(r.text)
    return None


def show_version(cmd):  # pylint: disable=unused-argument
    from azure.cli.core.util import get_az_version_json
    versions = get_az_version_json()
    return versions



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from azure.cli.core.commands import LongRunningOperation, CliCommandType
from ._client_factory import iot_hub_service_factory
from ._client_factory import iot_service_provisioning_factory
from ._client_factory import iot_central_service_factory

JOB_DEPRECATION_INFO = 'IoT Extension (azure-cli-iot-ext) Job commands'


class PolicyUpdateResultTransform(LongRunningOperation):  # pylint: disable=too-few-public-methods
    def __call__(self, poller):
        result = super(PolicyUpdateResultTransform, self).__call__(poller)
        return result.properties.authorization_policies


class EndpointUpdateResultTransform(LongRunningOperation):  # pylint: disable=too-few-public-methods
    def __call__(self, poller):
        result = super(EndpointUpdateResultTransform, self).__call__(poller)
        return result.properties.routing.endpoints


class RouteUpdateResultTransform(LongRunningOperation):  # pylint: disable=too-few-public-methods
    def __call__(self, poller):
        result = super(RouteUpdateResultTransform, self).__call__(poller)
        return result.properties.routing.routes


# Deleting IoT Hub is a long-running operation. Due to API implementation issue, 404 error will be thrown during
# deletion of an IoT Hub.
# This is a work around to suppress the 404 error. It should be removed after API is fixed.
class HubDeleteResultTransform(LongRunningOperation):  # pylint: disable=too-few-public-methods
    def __call__(self, poller):
        from azure.cli.core.util import CLIError
        try:
            super(HubDeleteResultTransform, self).__call__(poller)
        except CLIError as e:
            if 'not found' not in str(e):
                raise e


def load_command_table(self, _):  # pylint: disable=too-many-statements

    update_custom_util = CliCommandType(operations_tmpl='azure.cli.command_modules.iot.custom#{}')

    iot_central_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.iotcentral.operations#IoTCentaralOperations.{}'
    )

    # iot dps commands
    with self.command_group('iot dps', client_factory=iot_service_provisioning_factory) as g:
        g.custom_command('list', 'iot_dps_list')
        g.custom_show_command('show', 'iot_dps_get')
        g.custom_command('create', 'iot_dps_create')
        g.custom_command('delete', 'iot_dps_delete')
        g.generic_update_command('update', getter_name='iot_dps_get', setter_name='iot_dps_update',
                                 command_type=update_custom_util)

    # iot dps access-policy commands
    with self.command_group('iot dps access-policy', client_factory=iot_service_provisioning_factory) as g:
        g.custom_command('list', 'iot_dps_access_policy_list')
        g.custom_show_command('show', 'iot_dps_access_policy_get')
        g.custom_command('create', 'iot_dps_access_policy_create', supports_no_wait=True)
        g.custom_command('update', 'iot_dps_access_policy_update', supports_no_wait=True)
        g.custom_command('delete', 'iot_dps_access_policy_delete', supports_no_wait=True)

    # iot dps linked-hub commands
    with self.command_group('iot dps linked-hub', client_factory=iot_service_provisioning_factory) as g:
        g.custom_command('list', 'iot_dps_linked_hub_list')
        g.custom_show_command('show', 'iot_dps_linked_hub_get')
        g.custom_command('create', 'iot_dps_linked_hub_create', supports_no_wait=True)
        g.custom_command('update', 'iot_dps_linked_hub_update', supports_no_wait=True)
        g.custom_command('delete', 'iot_dps_linked_hub_delete', supports_no_wait=True)

    # iot dps certificate commands
    with self.command_group('iot dps certificate', client_factory=iot_service_provisioning_factory) as g:
        g.custom_command('list', 'iot_dps_certificate_list')
        g.custom_show_command('show', 'iot_dps_certificate_get')
        g.custom_command('create', 'iot_dps_certificate_create')
        g.custom_command('delete', 'iot_dps_certificate_delete')
        g.custom_command('generate-verification-code', 'iot_dps_certificate_gen_code')
        g.custom_command('verify', 'iot_dps_certificate_verify')
        g.custom_command('update', 'iot_dps_certificate_update')

    # iot hub certificate commands
    with self.command_group('iot hub certificate', client_factory=iot_hub_service_factory) as g:
        g.custom_command('list', 'iot_hub_certificate_list')
        g.custom_show_command('show', 'iot_hub_certificate_get')
        g.custom_command('create', 'iot_hub_certificate_create')
        g.custom_command('delete', 'iot_hub_certificate_delete')
        g.custom_command('generate-verification-code', 'iot_hub_certificate_gen_code')
        g.custom_command('verify', 'iot_hub_certificate_verify')
        g.custom_command('update', 'iot_hub_certificate_update')

    # iot hub commands
    with self.command_group('iot hub', client_factory=iot_hub_service_factory) as g:
        g.custom_command('create', 'iot_hub_create')
        g.custom_command('list', 'iot_hub_list')
        g.custom_command('show-connection-string', 'iot_hub_show_connection_string')
        g.custom_show_command('show', 'iot_hub_get')
        g.generic_update_command('update', getter_name='iot_hub_get', setter_name='iot_hub_update',
                                 command_type=update_custom_util, custom_func_name='update_iot_hub_custom')
        g.custom_command('delete', 'iot_hub_delete', transform=HubDeleteResultTransform(self.cli_ctx))
        g.custom_command('list-skus', 'iot_hub_sku_list')
        g.custom_command('show-quota-metrics', 'iot_hub_get_quota_metrics')
        g.custom_command('show-stats', 'iot_hub_get_stats')
        g.custom_command('manual-failover', 'iot_hub_manual_failover', supports_no_wait=True)

    # iot hub consumer group commands
    with self.command_group('iot hub consumer-group', client_factory=iot_hub_service_factory) as g:
        g.custom_command('create', 'iot_hub_consumer_group_create')
        g.custom_command('list', 'iot_hub_consumer_group_list')
        g.custom_command('show', 'iot_hub_consumer_group_get')
        g.custom_command('delete', 'iot_hub_consumer_group_delete')

    # iot hub policy commands
    with self.command_group('iot hub policy', client_factory=iot_hub_service_factory) as g:
        g.custom_command('list', 'iot_hub_policy_list')
        g.custom_command('show', 'iot_hub_policy_get')
        g.custom_command('create', 'iot_hub_policy_create', transform=PolicyUpdateResultTransform(self.cli_ctx))
        g.custom_command('delete', 'iot_hub_policy_delete', transform=PolicyUpdateResultTransform(self.cli_ctx))
        g.custom_command('renew-key', 'iot_hub_policy_key_renew', supports_no_wait=True)

    # iot hub routing endpoint commands
    with self.command_group('iot hub routing-endpoint', client_factory=iot_hub_service_factory) as g:
        g.custom_command('create', 'iot_hub_routing_endpoint_create',
                         transform=EndpointUpdateResultTransform(self.cli_ctx))
        g.custom_command('show', 'iot_hub_routing_endpoint_show')
        g.custom_command('list', 'iot_hub_routing_endpoint_list')
        g.custom_command('delete', 'iot_hub_routing_endpoint_delete',
                         transform=EndpointUpdateResultTransform(self.cli_ctx))

    # iot hub message enrichment commands
    with self.command_group('iot hub message-enrichment', client_factory=iot_hub_service_factory,
                            min_api="2019-07-01-preview") as g:
        g.custom_command('create', 'iot_message_enrichment_create')
        g.custom_command('list', 'iot_message_enrichment_list')
        g.custom_command('delete', 'iot_message_enrichment_delete')
        g.custom_command('update', 'iot_message_enrichment_update')

    # iot hub route commands
    with self.command_group('iot hub route', client_factory=iot_hub_service_factory) as g:
        g.custom_command('create', 'iot_hub_route_create', transform=RouteUpdateResultTransform(self.cli_ctx))
        g.custom_command('show', 'iot_hub_route_show')
        g.custom_command('list', 'iot_hub_route_list')
        g.custom_command('delete', 'iot_hub_route_delete', transform=RouteUpdateResultTransform(self.cli_ctx))
        g.custom_command('update', 'iot_hub_route_update', transform=RouteUpdateResultTransform(self.cli_ctx))
        g.custom_command('test', 'iot_hub_route_test')

    # iot hub device stream commands
    with self.command_group('iot hub devicestream', client_factory=iot_hub_service_factory,
                            min_api="2019-07-01-preview") as g:
        g.custom_command('show', 'iot_hub_devicestream_show')

    with self.command_group('iot central app', iot_central_sdk, client_factory=iot_central_service_factory,
                            is_preview=True) as g:
        g.custom_command('create', 'iot_central_app_create')
        g.custom_command('list', 'iot_central_app_list')
        g.custom_show_command('show', 'iot_central_app_get')
        g.generic_update_command('update', getter_name='iot_central_app_get',
                                 setter_name='iot_central_app_update', command_type=update_custom_util)
        g.custom_command('delete', 'iot_central_app_delete', confirmation=True)



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

from azure.cli.core.commands import CliCommandType
from ._client_factory import (cf_security_tasks,
                              cf_security_alerts,
                              cf_security_settings,
                              cf_security_contacts,
                              cf_security_auto_provisioning_settings,
                              cf_security_discovered_security_solutions,
                              cf_security_external_security_solutions,
                              cf_security_jit_network_access_policies,
                              cf_security_locations,
                              cf_security_pricings,
                              cf_security_topology,
                              cf_security_workspace_settings,
                              cf_security_advanced_threat_protection,
                              cf_security_assessment,
                              cf_security_assessment_metadata,
                              cf_security_sub_assessment,
                              cf_security_adaptive_application_controls,
                              cf_security_adaptive_network_hardenings,
                              cf_security_allowed_connections,
                              cf_security_iot_solution,
                              cf_security_iot_analytics,
                              cf_security_iot_alerts,
                              cf_security_iot_recommendations,
                              cf_security_regulatory_compliance_standards,
                              cf_security_regulatory_compliance_control,
                              cf_security_regulatory_compliance_assessment)


# pylint: disable=line-too-long
# pylint: disable=too-many-statements
# pylint: disable=too-many-locals
def load_command_table(self, _):

    security_regulatory_compliance_standards_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.security.operations#RegulatoryComplianceStandardsOperations.{}',
        client_factory=cf_security_regulatory_compliance_standards
    )

    security_regulatory_compliance_controls_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.security.operations#RegulatoryComplianceControlsOperations.{}',
        client_factory=cf_security_regulatory_compliance_control
    )

    security_regulatory_compliance_assessment_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.security.operations#RegulatoryComplianceAssessmentsOperations.{}',
        client_factory=cf_security_regulatory_compliance_assessment
    )

    security_tasks_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.security.operations#TasksOperations.{}',
        client_factory=cf_security_tasks,
        operation_group='security_tasks'
    )

    security_alerts_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.security.operations#AlertsOperations.{}',
        client_factory=cf_security_alerts,
        operation_group='security_alerts'
    )

    security_settings_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.security.operations#SettingsOperations.{}',
        client_factory=cf_security_settings,
        operation_group='security_settings'
    )

    security_contacts_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.security.operations#SecurityContactsOperations.{}',
        client_factory=cf_security_contacts,
        operation_group='security_contacts'
    )

    security_auto_provisioning_settings_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.security.operations#AutoProvisioningSettingsOperations.{}',
        client_factory=cf_security_auto_provisioning_settings,
        operation_group='security_auto_provisioning_settings'
    )

    security_discovered_security_solutions_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.security.operations#DiscoveredSecuritySolutionsOperations.{}',
        client_factory=cf_security_discovered_security_solutions,
        operation_group='security_discovered_security_solutions'
    )

    security_external_security_solutions_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.security.operations#ExternalSecuritySolutionsOperations.{}',
        client_factory=cf_security_external_security_solutions,
        operation_group='security_external_security_solutions'
    )

    security_jit_network_access_policies_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.security.operations#JitNetworkAccessPoliciesOperations.{}',
        client_factory=cf_security_jit_network_access_policies,
        operation_group='security_jit_network_access_policies'
    )

    security_locations_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.security.operations#LocationsOperations.{}',
        client_factory=cf_security_locations,
        operation_group='security_locations'
    )

    security_pricings_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.security.operations#PricingsOperations.{}',
        client_factory=cf_security_pricings,
        operation_group='security_pricings'
    )

    security_topology_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.security.operations#TopologyOperations.{}',
        client_factory=cf_security_topology,
        operation_group='security_topology'
    )

    security_workspace_settings_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.security.operations#WorkspaceSettingsOperations.{}',
        client_factory=cf_security_workspace_settings,
        operation_group='security_workspace_settings'
    )

    security_advanced_threat_protection_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.security.operations#AdvancedThreatProtectionOperations.{}',
        client_factory=cf_security_advanced_threat_protection
    )

    security_assessment_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.security.operations#AssessmentsOperations.{}',
        client_factory=cf_security_assessment
    )

    security_assessment_metadata_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.security.operations#AssessmentMetadataOperations.{}',
        client_factory=cf_security_assessment_metadata
    )

    security_sub_assessment_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.security.operations#SubAssessmentsOperations.{}',
        client_factory=cf_security_sub_assessment
    )

    security_adaptive_application_controls_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.security.operations#AdaptiveApplicationControlsOperations.{}',
        client_factory=cf_security_adaptive_application_controls,
        operation_group='cf_security_adaptive_application_controls'
    )

    security_adaptive_network_hardenings_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.security.operations#AdaptiveNetworkhardeningsOperations.{}',
        client_factory=cf_security_adaptive_network_hardenings,
        operation_group='security_adaptive_network_hardenings'
    )

    security_allowed_connections_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.security.operations#AllowedConnectionsOperations.{}',
        client_factory=cf_security_allowed_connections,
        operation_group='security_allowed_connections'
    )

    security_iot_solution_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.security.operations#IotSolutionOperations.{}',
        client_factory=cf_security_iot_solution
    )

    security_iot_analytics_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.security.operations#IotAnalyticsOperations.{}',
        client_factory=cf_security_iot_analytics
    )

    security_iot_alerts_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.security.operations#IotAlertsOperations.{}',
        client_factory=cf_security_iot_alerts
    )

    security_iot_recommendations_sdk = CliCommandType(
        operations_tmpl='azure.mgmt.security.operations#IotRecommendationsOperations.{}',
        client_factory=cf_security_iot_recommendations
    )

    with self.command_group('security regulatory-compliance-standards',
                            security_regulatory_compliance_standards_sdk,
                            client_factory=cf_security_regulatory_compliance_standards) as g:
        g.custom_command('list', 'list_regulatory_compliance_standards')
        g.custom_command('show', 'get_regulatory_compliance_standard')

    with self.command_group('security regulatory-compliance-controls',
                            security_regulatory_compliance_controls_sdk,
                            client_factory=cf_security_regulatory_compliance_control) as g:
        g.custom_command('list', 'list_regulatory_compliance_controls')
        g.custom_command('show', 'get_regulatory_compliance_control')

    with self.command_group('security regulatory-compliance-assessments',
                            security_regulatory_compliance_assessment_sdk,
                            client_factory=cf_security_regulatory_compliance_assessment) as g:
        g.custom_command('list', 'list_regulatory_compliance_assessments')
        g.custom_command('show', 'get_regulatory_compliance_assessment')

    with self.command_group('security task',
                            security_tasks_sdk,
                            client_factory=cf_security_tasks) as g:
        g.custom_command('list', 'list_security_tasks')
        g.custom_command('show', 'get_security_task')

    with self.command_group('security atp storage',
                            security_advanced_threat_protection_sdk,
                            client_factory=cf_security_advanced_threat_protection) as g:
        g.custom_command('show', 'get_atp_setting')
        g.custom_command('update', 'update_atp_setting')

    with self.command_group('security alert',
                            security_alerts_sdk,
                            client_factory=cf_security_alerts) as g:
        g.custom_command('list', 'list_security_alerts')
        g.custom_command('show', 'get_security_alert')
        g.custom_command('update', 'update_security_alert')

    with self.command_group('security setting',
                            security_settings_sdk,
                            client_factory=cf_security_settings) as g:
        g.custom_command('list', 'list_security_settings')
        g.custom_command('show', 'get_security_setting')

    with self.command_group('security contact',
                            security_contacts_sdk,
                            client_factory=cf_security_contacts) as g:
        g.custom_command('list', 'list_security_contacts')
        g.custom_command('show', 'get_security_contact')
        g.custom_command('create', 'create_security_contact')
        g.custom_command('delete', 'delete_security_contact')

    with self.command_group('security auto-provisioning-setting',
                            security_auto_provisioning_settings_sdk,
                            client_factory=cf_security_auto_provisioning_settings) as g:
        g.custom_command('list', 'list_security_auto_provisioning_settings')
        g.custom_command('show', 'get_security_auto_provisioning_setting')
        g.custom_command('update', 'update_security_auto_provisioning_setting')

    with self.command_group('security discovered-security-solution',
                            security_discovered_security_solutions_sdk,
                            client_factory=cf_security_discovered_security_solutions) as g:
        g.custom_command('list', 'list_security_discovered_security_solutions')
        g.custom_command('show', 'get_security_discovered_security_solution')

    with self.command_group('security external-security-solution',
                            security_external_security_solutions_sdk,
                            client_factory=cf_security_external_security_solutions) as g:
        g.custom_command('list', 'list_security_external_security_solutions')
        g.custom_command('show', 'get_security_external_security_solution')

    with self.command_group('security jit-policy',
                            security_jit_network_access_policies_sdk,
                            client_factory=cf_security_jit_network_access_policies) as g:
        g.custom_command('list', 'list_security_jit_network_access_policies')
        g.custom_command('show', 'get_security_jit_network_access_policy')

    with self.command_group('security location',
                            security_locations_sdk,
                            client_factory=cf_security_locations) as g:
        g.custom_command('list', 'list_security_locations')
        g.custom_command('show', 'get_security_location')

    with self.command_group('security pricing',
                            security_pricings_sdk,
                            client_factory=cf_security_pricings) as g:
        g.custom_command('list', 'list_security_pricings')
        g.custom_command('show', 'get_security_pricing')
        g.custom_command('create', 'create_security_pricing')

    with self.command_group('security topology',
                            security_topology_sdk,
                            client_factory=cf_security_topology) as g:
        g.custom_command('list', 'list_security_topology')
        g.custom_command('show', 'get_security_topology')

    with self.command_group('security workspace-setting',
                            security_workspace_settings_sdk,
                            client_factory=cf_security_workspace_settings) as g:
        g.custom_command('list', 'list_security_workspace_settings')
        g.custom_command('show', 'get_security_workspace_setting')
        g.custom_command('create', 'create_security_workspace_setting')
        g.custom_command('delete', 'delete_security_workspace_setting')

    with self.command_group('security assessment',
                            security_assessment_sdk,
                            client_factory=cf_security_assessment) as g:
        g.custom_command('list', 'list_security_assessments')
        g.custom_command('show', 'get_security_assessment')
        g.custom_command('create', 'create_security_assessment')
        g.custom_command('delete', 'delete_security_assessment')

    with self.command_group('security assessment-metadata',
                            security_assessment_metadata_sdk,
                            client_factory=cf_security_assessment_metadata) as g:
        g.custom_command('list', 'list_security_assessment_metadata')
        g.custom_command('show', 'get_security_assessment_metadata')
        g.custom_command('create', 'create_security_assessment_metadata')
        g.custom_command('delete', 'delete_security_assessment_metadata')

    with self.command_group('security sub-assessment',
                            security_sub_assessment_sdk,
                            client_factory=cf_security_sub_assessment) as g:
        g.custom_command('list', 'list_security_sub_assessments')
        g.custom_command('show', 'get_security_sub_assessment')

    with self.command_group('security adaptive-application-controls',
                            security_adaptive_application_controls_sdk,
                            client_factory=cf_security_adaptive_application_controls) as g:
        g.custom_command('list', 'list_security_adaptive_application_controls')
        g.custom_command('show', 'get_security_adaptive_application_controls')

    with self.command_group('security adaptive_network_hardenings',
                            security_adaptive_network_hardenings_sdk,
                            client_factory=cf_security_adaptive_network_hardenings) as g:
        g.custom_command('show', 'get_security_adaptive_network_hardenings')
        g.custom_command('list', 'list_security_adaptive_network_hardenings')

    with self.command_group('security allowed_connections',
                            security_allowed_connections_sdk,
                            client_factory=cf_security_allowed_connections) as g:
        g.custom_command('list', 'list_security_allowed_connections')
        g.custom_command('show', 'get_security_allowed_connections')

    with self.command_group('security iot-solution',
                            security_iot_solution_sdk,
                            client_factory=cf_security_iot_solution) as g:
        g.custom_command('list', 'list_security_iot_solution')
        g.custom_show_command('show', 'show_security_iot_solution')
        g.custom_command('create', 'create_security_iot_solution')
        g.custom_command('delete', 'delete_security_iot_solution')
        g.custom_command('update', 'update_security_iot_solution')

    with self.command_group('security iot-analytics',
                            security_iot_analytics_sdk,
                            client_factory=cf_security_iot_analytics) as g:
        g.custom_command('list', 'list_security_iot_analytics')
        g.custom_command('show', 'show_security_iot_analytics')

    with self.command_group('security iot-alerts',
                            security_iot_alerts_sdk,
                            client_factory=cf_security_iot_alerts) as g:
        g.custom_command('list', 'list_security_iot_alerts')
        g.custom_show_command('show', 'show_security_iot_alerts')
        g.custom_command('delete', 'dismiss_security_iot_alerts')

    with self.command_group('security iot-recommendations',
                            security_iot_recommendations_sdk,
                            client_factory=cf_security_iot_recommendations) as g:
        g.custom_command('list', 'list_security_iot_recommendations')
        g.custom_show_command('show', 'show_security_iot_recommendations')

    with self.command_group('security', is_preview=True):
        pass



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

# pylint: disable=line-too-long

import json
import requests
import azure.cli.command_modules.appconfig._azconfig.keyvalue_iterable as iterable
import azure.cli.command_modules.appconfig._azconfig.exceptions as exceptions
import azure.cli.command_modules.appconfig._azconfig.constants as constants
import azure.cli.command_modules.appconfig._azconfig.utils as utils
import azure.cli.command_modules.appconfig._azconfig.models as models
import azure.cli.command_modules.appconfig._azconfig.mapper as mapper
import azure.cli.command_modules.appconfig._azconfig.request_message as request_message
import azure.cli.command_modules.appconfig._azconfig.request_handler as handler


class AzconfigClient:
    """Represents an azconfig client.

    Provides a client-side logical representation of the Azure config service.
    This client is used to configure and execute requests against the
    service.

    The service client encapsulates the endpoint and credentials used to access
    the Azure config service.
    """

    def __init__(self, connection_string, client_options=None):
        """
        :param string connection_params:
            Contains 'endpoint', 'id' and 'secret', where id and secret are credentials
            used to create the client.
        :param ClientOptions client_options:
            Optional parameter to customize AzconfigClient

        """
        self.connection_string = connection_string
        self._request_session = requests.Session()

        self._client_options = models.ClientOptions(
        ) if client_options is None else client_options

        self._default_headers = {
            constants.HttpHeaders.UserAgent: self._client_options.user_agent
        }

        self._request_handler = handler.RequestHandler(
            connection_string, self._client_options)

    def add_keyvalue(self, keyvalue, modify_options=None):
        """ Adds a new key-value to a configuration store.

        :param KeyValue keyvalue:
            The key-value to add.
        :param dict custom_headers:
            Headers that will be added to the request
        :param ModifyKeyValueOptions modify_options:
            Optional parameter to set keyvalue modification options

        :return:
            The key-value that was added to the configuration store.
        :rtype:
            KeyValue

        :raises ValueError: If the keyvalue entry alreay exists.

        """
        if modify_options is None:
            modify_options = models.ModifyKeyValueOptions()

        key, label = utils.unescape_encode_key_and_label(keyvalue.key, keyvalue.label)
        body_content = {
            "content_type": keyvalue.content_type,
            "value": keyvalue.value,
            "tags": keyvalue.tags
        }

        return self.__write_key(key,
                                label,
                                body_content,
                                modify_options,
                                if_match_etag=None,
                                if_none_match_etag='*')

    def set_keyvalue(self, keyvalue, modify_options=None):
        """ Sets a key-value's properties within a configuration store.
        If the key-value does not exist it will be created.

        :param KeyValue keyvalue:
            The key-value to set.
        :param ModifyKeyValueOptions modify_options:
            Optional parameter to set keyvalue modification options

        :return:
            The key-value that was set in the configuration store.
        :rtype:
            KeyValue

        """
        if modify_options is None:
            modify_options = models.ModifyKeyValueOptions()

        key, label = utils.unescape_encode_key_and_label(keyvalue.key, keyvalue.label)
        body_content = {
            "content_type": keyvalue.content_type,
            "value": keyvalue.value,
            "tags": keyvalue.tags
        }
        return self.__write_key(key,
                                label,
                                body_content,
                                modify_options)

    def update_keyvalue(self, keyvalue, modify_options=None):
        """ Updates a key-value that was retrieved from a configuration store.
        The ETag property is used to ensure that no external changes to the key-value have
        occurred in the configuration store since the key-value was retrieved.

        :param KeyValue keyvalue:
            The key-value to update.
        :param ModifyKeyValueOptions modify_options:
            Optional parameter to set keyvalue modification options

        :return:
            The updated key-value.
        :rtype:
            KeyValue

        :raises ValueError: If the keyvalue entry has been modified and etag mismatches.

        """
        if modify_options is None:
            modify_options = models.ModifyKeyValueOptions()

        if keyvalue.etag is None:
            raise ValueError("Etag of the keyvalue cannot be null")

        key, label = utils.unescape_encode_key_and_label(keyvalue.key, keyvalue.label)
        body_content = {
            "content_type": keyvalue.content_type,
            "value": keyvalue.value,
            "tags": keyvalue.tags
        }
        return self.__write_key(key,
                                label,
                                body_content,
                                modify_options,
                                keyvalue.etag)

    def delete_keyvalue_by_key_label(self, key, label=None, modify_options=None):
        """ Deletes a key-value from a configuration store.

        :param str key:
            The key of the key-value that should be deleted.
        :param str label:
            The label of the key-value that should be deleted.
        :param ModifyKeyValueOptions modify_options:
            Optional parameter to set keyvalue modification options

        :return:
            The deleted key-value if found, otherwise null.
        :rtype:
            KeyValue

        """
        if modify_options is None:
            modify_options = models.ModifyKeyValueOptions()

        key, label = utils.unescape_encode_key_and_label(key, label)
        query_url = '/kv/{}?label={}'.format(key, '' if label is None else label)
        query_url = self.__append_api_version(query_url)

        endpoint = utils.get_endpoint_from_connection_string(
            self.connection_string)
        url = 'https://{}{}'.format(endpoint, query_url)

        custom_headers = self.__configure_request_ids(modify_options)
        custom_headers.update(self._default_headers)

        headers = utils.generate_request_header(method=constants.HttpMethods.Delete,
                                                custom_headers=custom_headers,
                                                datetime_=None,
                                                if_match_etag=None)
        response = self._request_handler.execute(request_message.RequestMessage(
            constants.HttpMethods.Delete, headers, url, ''), self._request_session)

        if response.status_code == constants.StatusCodes.OK:
            return mapper.map_json_to_keyvalue(response.json())
        if response.status_code == constants.StatusCodes.NO_CONTENT:
            return None

        raise exceptions.HTTPException(response.status_code, response.reason,
                                       response.headers, response.content)

    def delete_keyvalue(self, keyvalue, modify_options=None):
        """ Deletes a key-value from a configuration store.
        The ETag property is used to ensure that no external changes to the key-value have occurred in the configuration store since the key-value was retrieved.

        :param str keyvalue:
            The key-value to delete.
        :param ModifyKeyValueOptions modify_options:
            Optional parameter to set keyvalue modification options

        :return:
            The deleted key-value.
        :rtype:
            KeyValue

        :raises ValueError: If the key-value entry has been modified and etag mismatches.

        """
        if modify_options is None:
            modify_options = models.ModifyKeyValueOptions()

        if keyvalue.etag is None:
            raise ValueError("Etag of the keyvalue cannot be null")

        key, label = utils.unescape_encode_key_and_label(keyvalue.key, keyvalue.label)
        query_url = '/kv/{}?label={}'.format(key, '' if label is None else label)
        query_url = self.__append_api_version(query_url)

        endpoint = utils.get_endpoint_from_connection_string(
            self.connection_string)
        url = 'https://{}{}'.format(endpoint, query_url)

        custom_headers = self.__configure_request_ids(modify_options)
        custom_headers.update(self._default_headers)

        headers = utils.generate_request_header(method=constants.HttpMethods.Delete,
                                                custom_headers=custom_headers,
                                                datetime_=None,
                                                if_match_etag=keyvalue.etag)
        response = self._request_handler.execute(request_message.RequestMessage(
            constants.HttpMethods.Delete, headers, url, ''), self._request_session)

        if response.status_code == constants.StatusCodes.OK:
            return mapper.map_json_to_keyvalue(response.json())
        if response.status_code == constants.StatusCodes.PRECONDITION_FAILED:
            raise ValueError('The keyvalue entry has been modified.')

        raise exceptions.HTTPException(response.status_code, response.reason,
                                       response.headers, response.content)

    def get_keyvalue(self, key, query_options=None):
        """Retrieves a key-value with the specified key, taking into account the constraints of the key-value query options.

        :param str key:
            The key of the key-value to retrieved.
        :param QueryKeyValueOptions query_options:
            Parameters used to modify which key-value is retrieved.

        :return:
            The key-value if found, otherwise null
        :rtype:
            KeyValue
        """

        if query_options is None:
            query_options = models.QueryKeyValueOptions()

        return self.__query_key(key, query_options)

    def get_keyvalues(self, query_options=None):
        """Returns an iterable object which allows the caller to iterate and retrieve key-values.

        :param QueryKeyValueCollectionOptions query_options:
            Parameters used to modify the set of key-values that are retrieved.

        :return:
            An iterable of key-values if found, otherwise an empty one                                                                                                                                                                                                                                                                                                                                                                                                                                                                     .
        :rtype: KeyValueIterable

        """
        if query_options is None:
            query_options = models.QueryKeyValueCollectionOptions

        return iterable.KeyValueIterable(self, query_options, self.__query_keys)

    def read_keyvalue_revisions(self, query_options=None):
        """Returns an iterable object which allows the caller to asynchronously iterate and retrieve revisions.

        :param QueryKeyValueCollectionOptions query_options:
            Parameters used to modify the set of revisions that are retrieved.

        :return:
             An iterable of key-value revisions if found, otherwise an empty one
        :rtype: KeyValueIterable

        """
        if query_options is None:
            query_options = models.QueryKeyValueCollectionOptions

        return iterable.KeyValueIterable(self, query_options, self.__list_revision)

    def lock_keyvalue(self, keyvalue, modify_options=None):
        """Locks a key-value within a configuration store.

        :param KeyValue keyvalue:
            The key-value to be locked..
        :param ModifyKeyValueOptions modify_options:
            Optional parameter to set keyvalue modification options

        :return:
            The locked key-value if its ETag matches the ETag in the configuration store, otherwise null.
        :rtype: KeyValue

        """
        if modify_options is None:
            modify_options = models.ModifyKeyValueOptions()

        key, label = utils.unescape_encode_key_and_label(keyvalue.key, keyvalue.label)

        query_url = '/locks/{}'.format(key)
        query_url += '?label={}'.format('' if label is None else label)
        query_url = self.__append_api_version(query_url)

        endpoint = utils.get_endpoint_from_connection_string(
            self.connection_string)
        url = 'https://{}{}'.format(endpoint, query_url)

        custom_headers = self.__configure_request_ids(modify_options)
        custom_headers.update(self._default_headers)

        headers = utils.generate_request_header(method=constants.HttpMethods.Put,
                                                custom_headers=custom_headers,
                                                datetime_=None,
                                                if_match_etag=keyvalue.etag)
        response = self._request_handler.execute(request_message.RequestMessage(
            constants.HttpMethods.Put, headers, url, ''), self._request_session)

        if response.status_code == constants.StatusCodes.OK:
            return mapper.map_json_to_keyvalue(response.json())
        if response.status_code == constants.StatusCodes.NO_CONTENT:
            return None

        raise exceptions.HTTPException(response.status_code, response.reason,
                                       response.headers, response.content)

    def unlock_keyvalue(self, keyvalue, modify_options=None):
        """Unlocks a key-value within a configuration store.

        :param KeyValue keyvalue:
            The key-value to be unlocked.
        :param ModifyKeyValueOptions modify_options:
            Optional parameter to set keyvalue modification options

        :return:
            The unlocked key-value if its ETag matches the ETag in the configuration store, otherwise null.
        :rtype: KeyValue

        """
        if modify_options is None:
            modify_options = models.ModifyKeyValueOptions()

        key, label = utils.unescape_encode_key_and_label(keyvalue.key, keyvalue.label)

        query_url = '/locks/{}'.format(key)
        query_url += '?label={}'.format('' if label is None else label)
        query_url = self.__append_api_version(query_url)

        endpoint = utils.get_endpoint_from_connection_string(
            self.connection_string)
        url = 'https://{}{}'.format(endpoint, query_url)

        custom_headers = self.__configure_request_ids(modify_options)
        custom_headers.update(self._default_headers)

        headers = utils.generate_request_header(method=constants.HttpMethods.Delete,
                                                custom_headers=custom_headers,
                                                datetime_=None,
                                                if_match_etag=keyvalue.etag)
        response = self._request_handler.execute(request_message.RequestMessage(
            constants.HttpMethods.Delete, headers, url, ''), self._request_session)

        if response.status_code == constants.StatusCodes.OK:
            return mapper.map_json_to_keyvalue(response.json())
        if response.status_code == constants.StatusCodes.NO_CONTENT:
            return None

        raise exceptions.HTTPException(response.status_code, response.reason,
                                       response.headers, response.content)

    # pylint: disable=too-many-arguments

    def __write_key(self,
                    key,
                    label,
                    body_content,
                    modify_options,
                    if_match_etag=None,
                    if_none_match_etag=None):
        query_url = '/kv/{}?label={}'.format(key, '' if label is None else label)
        query_url = self.__append_api_version(query_url)

        endpoint = utils.get_endpoint_from_connection_string(
            self.connection_string)
        url = 'https://{}{}'.format(endpoint, query_url)

        custom_headers = self.__configure_request_ids(modify_options)
        custom_headers.update(self._default_headers)

        headers = utils.generate_request_header(method=constants.HttpMethods.Put,
                                                custom_headers=custom_headers,
                                                datetime_=None,
                                                if_match_etag=if_match_etag,
                                                if_none_match_etag=if_none_match_etag)

        response = self._request_handler.execute(request_message.RequestMessage(
            constants.HttpMethods.Put, headers, url, json.dumps(body_content)), self._request_session)

        if response.status_code == constants.StatusCodes.OK:
            return mapper.map_json_to_keyvalue(response.json())

        raise exceptions.HTTPException(response.status_code, response.reason,
                                       response.headers, response.content)

    def __list_revision(self, query_option, continuation_link):
        key, label = utils.unescape_encode_key_and_label(query_option.key_filter, query_option.label_filter)
        query_datetime = query_option.query_datetime
        query_fields = self.__construct_query_fields_to_string(
            query_option.fields)

        if continuation_link is None:
            query_url = '/revisions?key={}'.format('*' if key is None else key)
            query_url += '&label={}'.format('*'if label is None else label)
            query_url += '&fields={}'.format(query_fields)
            query_url = self.__append_api_version(query_url)
        else:
            query_url = self.__parse_link_header(continuation_link)
            if query_url is None:
                return [], None

        endpoint = utils.get_endpoint_from_connection_string(
            self.connection_string)
        url = 'https://{}{}'.format(endpoint, query_url)

        custom_headers = self.__configure_request_ids(query_option)
        custom_headers.update(self._default_headers)

        headers = utils.generate_request_header(method=constants.HttpMethods.Get,
                                                custom_headers=custom_headers,
                                                datetime_=query_datetime)
        response = self._request_handler.execute(request_message.RequestMessage(
            constants.HttpMethods.Get, headers, url, ''), self._request_session)

        if response.status_code == constants.StatusCodes.OK:
            if constants.HttpHeaders.Link in response.headers:
                return mapper.map_json_to_keyvalues(response.json()['items']), response.headers[constants.HttpHeaders.Link]
            return mapper.map_json_to_keyvalues(response.json()['items']), None

        raise exceptions.HTTPException(response.status_code, response.reason,
                                       response.headers, response.content)

    def __query_key(self, key, query_kv_option):
        key, label = utils.unescape_encode_key_and_label(key, query_kv_option.label)
        fields = self.__construct_query_fields_to_string(
            query_kv_option.fields)

        query_url = '/kv/{}?label={}'.format(key,
                                             label if label is not None else '')
        query_url += '&fields={}'.format('*' if fields is None else fields)
        query_url = self.__append_api_version(query_url)

        endpoint = utils.get_endpoint_from_connection_string(
            self.connection_string)
        url = 'https://{}{}'.format(endpoint, query_url)

        custom_headers = self.__configure_request_ids(query_kv_option)
        custom_headers.update(self._default_headers)

        headers = utils.generate_request_header(method=constants.HttpMethods.Get,
                                                custom_headers=custom_headers,
                                                datetime_=None if query_kv_option is None else query_kv_option.query_datetime)

        response = self._request_handler.execute(request_message.RequestMessage(
            constants.HttpMethods.Get, headers, url, ''), self._request_session)

        if response.status_code == constants.StatusCodes.OK:
            return mapper.map_json_to_keyvalue(response.json())
        if response.status_code == constants.StatusCodes.NOT_FOUND:
            return None

        raise exceptions.HTTPException(response.status_code, response.reason,
                                       response.headers, response.content)

    def __query_keys(self, query_kv_collection_option, continuation_link):
        key, label = utils.unescape_encode_key_and_label(query_kv_collection_option.key_filter, query_kv_collection_option.label_filter)

        query_datetime = query_kv_collection_option.query_datetime
        query_fields = self.__construct_query_fields_to_string(
            query_kv_collection_option.fields)

        if continuation_link is None:
            query_url = '/kv?key={}'.format('*' if key is None else key)
            query_url += '&label={}'.format('*' if label is None else label)
            query_url += '&fields={}'.format('*' if query_fields is None else query_fields)
            query_url = self.__append_api_version(query_url)
        else:
            query_url = self.__parse_link_header(continuation_link)
            if query_url is None:
                return [], None

        endpoint = utils.get_endpoint_from_connection_string(
            self.connection_string)
        url = 'https://{}{}'.format(endpoint, query_url)

        custom_headers = self.__configure_request_ids(
            query_kv_collection_option)
        custom_headers.update(self._default_headers)

        headers = utils.generate_request_header(method=constants.HttpMethods.Get,
                                                custom_headers=custom_headers,
                                                datetime_=query_datetime)
        response = self._request_handler.execute(request_message.RequestMessage(
            constants.HttpMethods.Get, headers, url, ''), self._request_session)

        if response.status_code == constants.StatusCodes.OK:
            if constants.HttpHeaders.Link in response.headers:
                return mapper.map_json_to_keyvalues(response.json()['items']), response.headers[constants.HttpHeaders.Link]

            return mapper.map_json_to_keyvalues(response.json()['items']), None

        raise exceptions.HTTPException(response.status_code, response.reason,
                                       response.headers, response.content)

    @staticmethod
    def __parse_link_header(link_header):
        # link header looks like "</kv?after=last>; rel="next", </kv?before=first>; rel="before""
        uri_start = 0
        for index, char in enumerate(link_header):
            if char == '<':
                uri_start = index
            elif char == '>':
                if str(link_header[index + 3: index + 13]) == '''rel="next"''':
                    return str(link_header[uri_start + 1: index])
        return None

    @staticmethod
    def __construct_query_fields_to_string(query_fields):
        query_fields_string = ''
        if query_fields is not None:
            for field in query_fields:
                if field == models.QueryFields.ALL:
                    query_fields = '*'
                    break
                query_fields_string += field.name.lower() + ","
        return query_fields_string[:-1]

    @staticmethod
    def __configure_request_ids(request_options):
        custom_headers = {}

        custom_headers[constants.HttpHeaders.ClientRequestId] = request_options.client_request_id
        custom_headers[constants.HttpHeaders.CorrelationRequestId] = request_options.correlation_request_id

        return custom_headers

    @staticmethod
    def __append_api_version(url):
        api_version = "&api-version=" if '?' in url else "?api-version="
        return url + api_version + constants.Versions.ApiVersion



# --------------------------------------------------------------------------------------------
# Copyright (c) Microsoft Corporation. All rights reserved.
# Licensed under the MIT License. See License.txt in the project root for license information.
# --------------------------------------------------------------------------------------------

# pylint: disable=too-few-public-methods


class RequestMessage:

    def __init__(self, method, headers, url, body):
        self.method = method
        self.headers = headers
        self.url = url
        self.body = body
